/*-
 * Copyright (C) 2001 Juergen Keil.
 * Copyright (C) 2001 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef	__GNUC__
#include <stdarg.h>
#define _SYS_VARARGS_H          /* avoid including sys/varargs.h with gcc */
#else
#include <sys/varargs.h>
#endif

#include <sys/modctl.h>
#include <sys/file.h>
#include <sys/stream.h>
#include <sys/conf.h>
#include <sys/audio.h>
#include "audioio_compat.h"
#include "audionull.h"
#include <sys/ddi.h>



static int audionull_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static int audionull_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);



static struct module_info audionull_minfo = {
    0xffff,		/* module ID number */
    "audnull",		/* module name */
    0,			/* min packet size */
    4096,		/* max packet size */
    48000*4*2,		/* high water mark: ~2 secs of audio */
    0x8000		/* low water mark */
};


static struct qinit audionull_rqueue = {
    audio_hlp_rput,
    audio_hlp_rsvc,
    audio_hlp_open,
    audio_hlp_close,
    NULL,
    &audionull_minfo
};

static struct qinit audionull_wqueue = {
    audio_hlp_wput,
    audio_hlp_wsvc,
    NULL,
    NULL,
    NULL,
    &audionull_minfo
};


static struct streamtab audionull_str_info = {
    &audionull_rqueue,
    &audionull_wqueue,
    NULL,
    NULL,
};

static struct cb_ops audionull_cb_ops = {
    nulldev,			/* cb_open */
    nulldev,			/* cb_close */
    nodev,			/* cb_strategy */
    nodev,			/* cb_print */
    nodev,			/* cb_dump */
    nodev,			/* cb_read */
    nodev,			/* cb_write */
    nodev,			/* cb_ioctl */
    nodev,			/* cb_devmap */
    nodev,			/* cb_mmap */
    nodev,			/* cb_segmap */
    nochpoll,			/* cb_chpoll */
    ddi_prop_op,		/* cb_prop_op */
    &audionull_str_info,	/* cb_str */
    D_NEW|D_MTSAFE|D_64BIT,	/* cb_flag */
    CB_REV,			/* cb_rev */
    nodev,			/* cb_aread */
    nodev,			/* cb_awrite */
};

static struct dev_ops  audionull_drv_ops = {
    DEVO_REV,
    0,

    audio_hlp_getinfo,
    nulldev,			/* devo_identify */
    nulldev,			/* devo_probe */
    audionull_attach,
    audionull_detach,
    nodev,			/* devo_reset */

    &audionull_cb_ops,
    NULL,			/* devo_bus_ops */
    NULL,			/* devo_power */
};

/*
 * Module linkage information for the kernel.
 */
static struct modldrv audionull_modldrv = {
    &mod_driverops,		/* Type of module.  This one is a driver */
    "null audio driver",
    &audionull_drv_ops		/* driver ops */
};


static struct modlinkage audionull_modlinkage = {
    MODREV_1,
    { &audionull_modldrv, NULL },
};


#if	DEBUG
#define	DPRINTF(...)	audionull_dprintf(__VA_ARGS__)
#else
#define	DPRINTF(...)	/**/
#endif

void
audionull_dprintf(char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    vcmn_err(CE_CONT, fmt, args);
    va_end(args);
}


static void *audionull_state;

int 
_init(void)
{
    int err;

    err = ddi_soft_state_init(&audionull_state,
			      sizeof(struct audionull_softstate), 1);
    if (err == 0) {
	if ((err = mod_install(&audionull_modlinkage)) != 0)
	    ddi_soft_state_fini(&audionull_state);
    }

    DPRINTF("_init, err=%d\n", err);

    return err;
}


int
_fini(void)
{
    int err;
    
    if ((err = mod_remove(&audionull_modlinkage)) == 0)
	ddi_soft_state_fini(&audionull_state);

    DPRINTF("_fini, err=%d\n", err);

    return err;
}


int
_info(struct modinfo *modinfop)
{
    return mod_info(&audionull_modlinkage, modinfop);
}


static int
audionull_cleanup(struct audionull_softstate *state)
{
    int err = DDI_SUCCESS;

    if (state->audio_handle) {
	err = audio_hlp_detach(state->audio_handle, DDI_DETACH);
    }

    return err;
}


static ah_ad_info_t audionull_ad_info;

static int
audionull_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
    int instance = ddi_get_instance(dip);
    struct audionull_softstate *state;
    int err = DDI_FAILURE;

    DPRINTF("attach, cmd=%d\n", cmd);

    switch (cmd) {
    case DDI_ATTACH:
	err = ddi_soft_state_zalloc(audionull_state, instance);
	if (err != DDI_SUCCESS)
	    break;

	state = ddi_get_soft_state(audionull_state, instance);
	
	state->audio_handle = audio_hlp_attach(dip, cmd, &audionull_ad_info);
	if (!state->audio_handle) {
	    cmn_err(CE_WARN, "failed to attach audio helper module");
	    audionull_cleanup(state);
    	    ddi_soft_state_free(audionull_state, instance);
	    err = DDI_FAILURE;
	    break;
	}

	audio_hlp_set_private(state->audio_handle, state);
	break;

    default:
	break;
    }

    DPRINTF("attach, err=%d\n", err);
    return err;
}


static int
audionull_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
    int instance = ddi_get_instance(dip);
    struct audionull_softstate *state = ddi_get_soft_state(audionull_state, instance);
    int err = DDI_SUCCESS;

    DPRINTF("detach, cmd=%d\n", cmd);

    switch (cmd) {
    case DDI_DETACH:

	if ((err = audionull_cleanup(state)) == DDI_SUCCESS) 
	   ddi_soft_state_free(audionull_state, instance);

	break;

    default:
	err = DDI_FAILURE;
	break;
    }

    DPRINTF("detach, err=%d\n", err);

    return err;
}


/*
 * Audio Driver functions (for ah_ad_entry_t)
 */

static int    
audionull_setup(audiohdl_t handle, int stream, int flags)
{
#if 0
    DPRINTF("audionull_setup, handle %p, stream %d, flags %x\n",
	    handle, stream, flags);
#else
    audio_hlp_log(handle, 4,
		  "audionull_setup, handle %p, stream %d, flags %x\n",
		  handle, stream, flags);
#endif
    return AUDIO_SUCCESS;
}


static void
audionull_teardown(audiohdl_t handle, int stream, int dir)
{
#if 0
    DPRINTF("audionull_teardown, handle %p, stream %d, dir %d\n",
	    handle, stream, dir);
#else
    audio_hlp_log(handle, 4,
		  "audionull_teardown, handle %p, stream %d, dir %d\n",
	    handle, stream, dir);
#endif
}


static int
audionull_set_config(audiohdl_t handle, int stream,
		     int command, int dir, int arg1, int arg2)
{
    DPRINTF("audionull_set_config, handle %p, stream %d, "
	    "command %d, dir %d, arg1 %d, arg2 %d\n",
	    handle, stream, command, dir, arg1, arg2);

    switch (command) {
    case AD_SET_GAIN:
	break;
    case AD_SET_GAIN_BAL:
	break;
    case AD_SET_PORT:
	break;
    case AD_SET_MONITOR_GAIN:
	break;
    case AD_OUTPUT_MUTE:
	break;
    case AD_MONO_MIC:
	break;
    case AD_MIC_BOOST:
	break;
    case AD_BASS_BOOST:
	break;
    case AD_MID_BOOST:
	break;
    case AD_TREBLE_BOOST:
	break;
    case AD_LOUDNESS:
	break;
    case AD_SET_DIAG_MODE:
	break;
    }
    
    /* do not fail, otherwise the driver will not unload */
    return AUDIO_SUCCESS;
}


static int
audionull_set_format(audiohdl_t handle, int stream,
		 int dir, int sample_rate, int channels,
		 int precision, int encoding)
{
    DPRINTF("audionull_set_format, handle %p, stream %d, dir %d, "
	    "sample_rate %d, channels %d, precision %d, encoding %d\n",
	    handle, stream, dir, sample_rate, channels, precision, encoding);

    /* do not fail, otherwise the driver will not unload */
    return AUDIO_SUCCESS;
}


static void audionull_play_chunk(void *arg)
{
    audiohdl_t handle = arg;
    struct audionull_softstate *state = audio_hlp_get_private(handle);

#define	CHUNK_SIZE	512
    unsigned short buf[CHUNK_SIZE];
    int len;

    DPRINTF("audionull_play_chunk timeout, handle %p, stream %d\n", handle, state->playing_stream);
    
    len = audio_hlp_get_audio(handle, buf, state->playing_stream, CHUNK_SIZE);

    DPRINTF(" read %d samples with audio_hlp_get_audio\n", len);
#if 0
    {
	int i;
	char line[132];
	for (i = 0; i < len+4; i++) {
	    if (i % 16 == 0) line[0] = 0;
	    sprintf(line + strlen(line), "%04x ", buf[i]);
	    if (i % 16 == 15) DPRINTF("%s\n", line);
	}
	if (i % 16) DPRINTF("%s\n", line);
    }
#endif

    if (len > 0 && ++state->num_timeouts < 100) {
	state->samples_played += len;
	state->play_timeout_id = timeout(audionull_play_chunk, handle, drv_usectohz(10000));
    } else {
	state->play_timeout_id = 0;
    }
}

static int
audionull_start_play(audiohdl_t handle, int stream)
{
    struct audionull_softstate *state = audio_hlp_get_private(handle);

    DPRINTF("audionull_start_play, handle %p, stream %d\n", handle, stream);
    
    if (!state->play_timeout_id) {
	state->playing_stream = stream;
	state->num_timeouts = 0;
	state->samples_played = 0;
	state->play_timeout_id = timeout(audionull_play_chunk, handle, drv_usectohz(10000));
    }

    return AUDIO_SUCCESS;
}


static void
audionull_pause_play(audiohdl_t handle, int stream)
{
    DPRINTF("audionull_pause_play, handle %p, stream %d\n", handle, stream);
}


static void
audionull_stop_play(audiohdl_t handle, int stream)
{
    struct audionull_softstate *state = audio_hlp_get_private(handle);

    DPRINTF("audionull_stop_play, handle %p, stream %d\n", handle, stream);

    DPRINTF("  played %d samples\n", state->samples_played);
}


static void audionull_record_chunk(void *arg)
{
    audiohdl_t handle = arg;
    struct audionull_softstate *state = audio_hlp_get_private(handle);

#define	CHUNK_SIZE	512
    unsigned short buf[CHUNK_SIZE];
    int i;

    DPRINTF("audionull_record_chunk timeout, handle %p, stream %d\n", handle, state->playing_stream);
    
    for (i = 0; i < CHUNK_SIZE; i++)
	buf[i] = i;
    audio_hlp_send_audio(handle, buf, state->playing_stream, CHUNK_SIZE);

    if (++state->num_record_timeouts < 100) {
	state->record_timeout_id = timeout(audionull_record_chunk, handle, drv_usectohz(10000));
    } else {
	state->record_timeout_id = 0;
    }
}


static int
audionull_start_record(audiohdl_t handle, int stream)
{
    struct audionull_softstate *state = audio_hlp_get_private(handle);

    DPRINTF("audionull_start_record, handle %p, stream %d\n", handle, stream);

    if (!state->record_timeout_id) {
	state->recording_stream = stream;
	state->num_record_timeouts = 0;
	state->record_timeout_id = timeout(audionull_record_chunk, handle, drv_usectohz(10000));
    }
    return AUDIO_SUCCESS;
}


static void
audionull_stop_record(audiohdl_t handle, int stream)
{
    struct audionull_softstate *state = audio_hlp_get_private(handle);

    DPRINTF("audionull_stop_record, handle %p, stream %d\n", handle, stream);

    if (state->record_timeout_id) {
	untimeout(state->record_timeout_id);
	state->record_timeout_id = 0;
    }
}


static int
audionull_ioctl(audiohdl_t handle, int channel,
		queue_t *q, mblk_t *mp, int *error)
{
    struct iocblk *iocp = (struct iocblk *)mp->b_rptr;

    DPRINTF("audionull_ioctl ioc_cmd=%x\n", iocp->ioc_cmd);

    switch (iocp->ioc_cmd) {
    default:
	return AD_NACK;
    }
}


static int
audionull_iocdata(audiohdl_t handle, int channel,
		  queue_t *q, mblk_t *mp, int *error)
{
    struct copyresp *copyresp = (struct copyresp *)mp->b_rptr;
    mblk_t *data_mp;
    caddr_t arg = (caddr_t) copyresp->cp_private;

    if (mp->b_cont == NULL) return AD_NACK;
    data_mp = mp->b_cont;

    DPRINTF("audionull_iocdata cp_cmd=%x, arg=%x\n", copyresp->cp_cmd, arg);

    switch (copyresp->cp_cmd) {
    default:
	return AD_NACK;
    }
}


static audio_info_t audionull_defaults = {
    {
	8000,		     	/* samples per second */
	AUDIO_CHANNELS_MONO,	/* number of interleaved channels */
	AUDIO_PRECISION_8,	/* bit-width of each sample */
	AUDIO_ENCODING_ULAW,	/* data encoding method */

	AUDIO_MID_GAIN,		/* gain level: 0 - 255 */
	AUDIO_SPEAKER,		/* selected I/O port */
	AUDIO_SPEAKER | AUDIO_HEADPHONE | AUDIO_LINE_OUT,
#ifdef	AUDIOIO_HAVE_MODPORTS
	AUDIO_SPEAKER | AUDIO_HEADPHONE | AUDIO_LINE_OUT,
#endif
	0,			/* Reserved for future use */

	0,			/* I/O buffer size */

	0,			/* number of samples converted */
	0,			/* End Of File counter (play only) */

	0,
	0,
	0,
	AUDIO_MID_BALANCE,	/* stereo channel balance */
    },
    {
	8000,		     	/* samples per second */
	AUDIO_CHANNELS_MONO,	/* number of interleaved channels */
	AUDIO_PRECISION_8,	/* bit-width of each sample */
	AUDIO_ENCODING_ULAW,	/* data encoding method */

	AUDIO_MID_GAIN,		/* gain level: 0 - 255 */
	AUDIO_MICROPHONE,		/* selected I/O port */
	AUDIO_MICROPHONE | AUDIO_LINE_IN | AUDIO_CD |
		AUDIO_AUX1_IN | AUDIO_AUX2_IN | AUDIO_CODEC_LOOPB_IN,
#ifdef	AUDIOIO_HAVE_MODPORTS
	AUDIO_MICROPHONE | AUDIO_LINE_IN | AUDIO_CD |
		AUDIO_AUX1_IN | AUDIO_AUX2_IN | AUDIO_CODEC_LOOPB_IN,
#endif
	0,			/* Reserved for future use */

	0,			/* I/O buffer size */

	0,			/* number of samples converted */
	0,			/* End Of File counter (play only) */

	0,
	0,
	0,
	AUDIO_MID_BALANCE,	/* stereo channel balance */
    },
    AUDIO_MIN_GAIN,		/* monitor_gain */
    0,				/* output_muted */
#ifdef	AUDIOIO_HAS_REFCNT
    0,				/* driver reference count */
#endif
    {0, 0},			/* reserved fields */
#ifdef	AUDIOIO_HAS_FEATURES
    AUDIO_HWFEATURE_DUPLEX | AUDIO_HWFEATURE_IN2OUT | \
    	AUDIO_HWFEATURE_PLAY | AUDIO_HWFEATURE_RECORD,
    AUDIO_SWFEATURE_MIXER,
    0,
#endif
};


/*
 * Audio Driver ops vector definition
 */
static ah_ad_entry_t audionull_entry = {
    audionull_setup,
    audionull_teardown,
    audionull_set_config,
    audionull_set_format,
    audionull_start_play,
    audionull_pause_play,
    audionull_stop_play,
    audionull_start_record,
    audionull_stop_record,
    audionull_ioctl,
    audionull_iocdata,
};


static audio_device_t audionull_dev_info = {
    "TOOLS,null",		/* name */
    "1.0",			/* version */
    "null"			/* config */
};


/* 0 term. list of sample rates (used as limits) */
static uint_t audionull_srs[] = {
    8000, 48000, 0
};


static uint_t audionull_chs[] = {
    AUDIO_CHANNELS_MONO,
    AUDIO_CHANNELS_STEREO,
    0,
};

static ah_ad_cap_comb_t audionull_cap_comb[] = {
    { AUDIO_PRECISION_16, AUDIO_ENCODING_LINEAR },  /* 16-bit signed PCM */
    { AUDIO_PRECISION_8,  AUDIO_ENCODING_LINEAR8 }, /* 8-bit unsigned PCM */
    { AUDIO_PRECISION_8,  AUDIO_ENCODING_ULAW },
    { 0 },
};


static ah_ad_info_t audionull_ad_info = {
    AD_VERS1,			/* Audio Driver interface version */
    &audionull_defaults,	/* Audio Driver audio_info_t struct */
    {				/* play capabilities */
	{ AD_SRS_FLAG_SR_LIMITS, audionull_srs },
	audionull_chs,
	0,
	2,
	16384,			/* buffer size */
    },
    {				/* record capabilities */
	{ AD_SRS_FLAG_SR_LIMITS, audionull_srs },
	audionull_chs,
	0,
	2,
	16384,			/* buffer size */
    },
    audionull_cap_comb,		/* list of play cap. combinations */
    audionull_cap_comb,		/* list of rec cap. combinations */
    &audionull_entry,		/* Audio Driver entry points */
    &audionull_dev_info,	/* device information */
};
