/*-
 * Copyright (C) 2001-2003 Juergen Keil.
 * Copyright (C) 2001-2003 TooLs GmbH.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by TooLs GmbH.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef	__GNUC__
#include <stdarg.h>
#define _SYS_VARARGS_H          /* avoid including sys/varargs.h with gcc */
#else
#include <sys/varargs.h>
#endif

#include <sys/modctl.h>
#include <sys/file.h>
#include <sys/stream.h>
#include <sys/conf.h>
#include <sys/pci.h>
#if	SUNOS >= 509
#include <sys/audio/ac97.h>
#else
#include <sys/ac97.h>
#endif
#include <sys/audio.h>
#include "audioio_compat.h"
#include "ac97_codec.h"
#include "audiovia8233.h"

#include <sys/ddi.h>



/*
 * hide gcc2 __attribute__ compiler extension from non-gcc compiler
 * and old gcc compiler
 */
#if     !defined(__GNUC__)
#define __attribute__(x)        /**/
#elif   __GNUC__ < 2
#define __attribute__(x)        /**/
#endif


static int audiov8233_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static int audiov8233_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);
static int audiov8233_fill_play_buffers(struct audiov8233_softstate *softstate);
static int audiov8233_prepare_record_buffers(struct audiov8233_softstate *softstate);

static audio_info_t audiov8233_defaults;


static struct module_info audiov8233_minfo = {
    0x8233,		/* module ID number */
    "audv8233",		/* module name */
    0,			/* min packet size */
    8192,		/* max packet size */
    48000*4*2,		/* high water mark: ~2 secs of audio */
    0x8000		/* low water mark */
};


static struct qinit audiov8233_rqueue = {
    audio_hlp_rput,
    audio_hlp_rsvc,
    audio_hlp_open,
    audio_hlp_close,
    NULL,
    &audiov8233_minfo
};

static struct qinit audiov8233_wqueue = {
    audio_hlp_wput,
    audio_hlp_wsvc,
    NULL,
    NULL,
    NULL,
    &audiov8233_minfo
};


static struct streamtab audiov8233_str_info = {
    &audiov8233_rqueue,
    &audiov8233_wqueue,
    NULL,
    NULL,
};

static struct cb_ops audiov8233_cb_ops = {
    nulldev,			/* cb_open */
    nulldev,			/* cb_close */
    nodev,			/* cb_strategy */
    nodev,			/* cb_print */
    nodev,			/* cb_dump */
    nodev,			/* cb_read */
    nodev,			/* cb_write */
    nodev,			/* cb_ioctl */
    nodev,			/* cb_devmap */
    nodev,			/* cb_mmap */
    nodev,			/* cb_segmap */
    nochpoll,			/* cb_chpoll */
    ddi_prop_op,		/* cb_prop_op */
    &audiov8233_str_info,	/* cb_str */
    D_NEW|D_MTSAFE|D_64BIT,	/* cb_flag */
    CB_REV,			/* cb_rev */
    nodev,			/* cb_aread */
    nodev,			/* cb_awrite */
};

static struct dev_ops  audiov8233_drv_ops = {
    DEVO_REV,
    0,

    audio_hlp_getinfo,
    nulldev,			/* devo_identify */
    nulldev,			/* devo_probe */
    audiov8233_attach,
    audiov8233_detach,
    nodev,			/* devo_reset */

    &audiov8233_cb_ops,
    NULL,			/* devo_bus_ops */
    NULL,			/* devo_power */
};

/*
 * Module linkage information for the kernel.
 */
static struct modldrv audiov8233_modldrv = {
    &mod_driverops,		/* Type of module.  This one is a driver */
    "via 8233 audio driver",
    &audiov8233_drv_ops		/* driver ops */
};


static struct modlinkage audiov8233_modlinkage = {
    MODREV_1,
    { &audiov8233_modldrv, NULL },
};


/*
 * This describes the VIA controller's DMA engines to the OS at hardware level
 * 
 * The first structure is for the "SGD table".
 * These must be in contiguous physical memory.
 *
 * The VIA8233 accesses SGD memory in 8 bytes quantities (=> minxfer == 8)
 * and the whole sgd table has a fixed size of 32*8 == 256 bytes. An sgd table
 * must be allocated in contiguous physical memory (=> s/g length == 1).
 * 
 * The typical DMA transfer is a burst of 8 longwords (=> DMA burstsizes
 * has 0x20 bit set; I added 0x10, 0x8 and 0x4, too)
 */
static ddi_dma_attr_t sgd_dma_attr = {
    DMA_ATTR_V0,
    0,				/* addr_low */
    0xffffffff,			/* addr_high */
    0x0000ffff,			/* DMA counter register */
    8,				/* DMA address alignment */
    				/*   sgd list must be aligned on a 8-byte boundary */
    0x3c,			/* DMA burstsizes ?? */
    8,				/* minxfer effective DMA size */
    0x0000ffff,			/* max DMA xfer size */
    0x00000fff,			/* segment boundary*/
    1,				/* s/g length */
    8,				/* granularity of device */
				/*   size of a sgd entry?? */
    0				/* flags */
};


/*
 * DMA attributes for sample data.  No need for physical contiguous
 * memory.
 *
 * The VIA8233 accesses sample data in longword (4 bytes) quantities
 * (=> minxfer == 4) and has a 16-bit word DMA counter (=> max DMA
 * xfer size == 0xffff * 2).  It does support scatter/gatter DMA
 * transfers from upto 32 memory locations (=> s/g length == 32).
 * 
 * The typical DMA transfer is a burst of 8 longwords (=> DMA burstsizes
 * has 0x20 bit set; I added 0x10, 0x8 and 0x4, too)
 */
static ddi_dma_attr_t dma_attr = {
    DMA_ATTR_V0,
    0,				/* addr_low */
    0xffffffff,			/* addr_high */
    0x00ffffff,			/* DMA counter register */
    2,				/* DMA address alignment */
    0x3c,			/* DMA burstsizes */
    4,				/* minxfer effective DMA size */
    0x00ffffff,			/* max DMA xfer size */
    0x00001fff,			/* segment boundary*/
    1/*V8233_NUM_SGD*/,		/* s/g length */
    4,				/* granularity of device */
				/*   (VIA AC97 dma operates on longwords) */
    0				/* flags */
};


/* This describes things like byteorder internally for the VIA audio controller */
static struct ddi_device_acc_attr dev_attr = {
    DDI_DEVICE_ATTR_V0,
    DDI_NEVERSWAP_ACC,
    DDI_STRICTORDER_ACC
};


/* available log levels */
#define	AUDV8233_LOG_L0	0		/* warnings, console & syslog buffer */
#define	AUDV8233_LOG_L1	1		/* errors, syslog buffer */
#define	AUDV8233_LOG_L2	2		/* recoverable errors, debug only */
#define	AUDV8233_LOG_L3	3		/* interesting data, debug only */
#define	AUDV8233_LOG_L4	4		/* tracing, debug only */


/* available log masks */
#define	AUDV8233M_AC97	 0x00000020	/* AC97 register access */
#define	AUDV8233M_CODEC	 0x00000010	/* CODEC IO */
#define	AUDV8233M_INTR	 0x00000008	/* interrupt handling */
#define	AUDV8233M_DRIVER 0x00000004	/* audio driver functions */
#define	AUDV8233M_ATTACH 0x00000002	/* card setup at attach time */
#define	AUDV8233M_MODS	 0x00000001	/* _init, _info, _fini */
#define	AUDV8233M_ALL	 (~0u)

uint_t audiov8233_errlevel = AUDV8233_LOG_L2;
uint_t audiov8233_errmask =  AUDV8233M_ALL;


/* logging for log levels 4..0 */
#if	DEBUG
#define	DPRINTF4(...)	audiov8233_dprintf4(__VA_ARGS__)
#define	DPRINTF3(...)	audiov8233_dprintf3(__VA_ARGS__)
#define	DPRINTF2(...)	audiov8233_dprintf2(__VA_ARGS__)
#else
#define	DPRINTF4(...)	/**/
#define	DPRINTF3(...)	/**/
#define	DPRINTF2(...)	/**/
#endif
#define	DPRINTF1(...)	audiov8233_dprintf1(__VA_ARGS__)
#define	DPRINTF0(...)	audiov8233_dprintf0(__VA_ARGS__)


static void
audiov8233_vprintf(uint_t level, uint_t mask, char *fmt, va_list args)
{
    if (level <= audiov8233_errlevel && (mask & audiov8233_errmask))
	vcmn_err(CE_CONT, fmt, args);
}


#ifdef	DEBUG
void __attribute__ ((format (__printf__, 2, 3)))
audiov8233_dprintf4(uint_t mask, char *fmt, ...)

{
    va_list args;
    va_start(args, fmt);
    audiov8233_vprintf(AUDV8233_LOG_L4, mask, fmt, args);
    va_end(args);
}

void __attribute__ ((format (__printf__, 2, 3)))
audiov8233_dprintf3(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audiov8233_vprintf(AUDV8233_LOG_L3, mask, fmt, args);
    va_end(args);
}

void __attribute__ ((format (__printf__, 2, 3)))
audiov8233_dprintf2(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audiov8233_vprintf(AUDV8233_LOG_L2, mask, fmt, args);
    va_end(args);
}
#endif

void __attribute__ ((format (__printf__, 2, 3)))
audiov8233_dprintf1(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audiov8233_vprintf(AUDV8233_LOG_L1, mask, fmt, args);
    va_end(args);
}

void __attribute__ ((format (__printf__, 2, 3)))
audiov8233_dprintf0(uint_t mask, char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    audiov8233_vprintf(AUDV8233_LOG_L0, mask, fmt, args);
    va_end(args);
}


/* 
 * Global softstate for multiple via 8233 audio controller instances
 */
static void *audiov8233_state;


/* 
 * Sun driver module initialization and finalization code 
 */
int 
_init(void)
{
    int err;

    err = ddi_soft_state_init(&audiov8233_state,
			      sizeof(struct audiov8233_softstate), 1);
    if (err == 0) {
	if ((err = mod_install(&audiov8233_modlinkage)) != 0)
	    ddi_soft_state_fini(&audiov8233_state);
    }

    DPRINTF4(AUDV8233M_MODS, "_init, err=%d\n", err);

    return err;
}


int
_fini(void)
{
    int err;
    
    if ((err = mod_remove(&audiov8233_modlinkage)) == 0)
	ddi_soft_state_fini(&audiov8233_state);

    DPRINTF4(AUDV8233M_MODS, "_fini, err=%d\n", err);

    return err;
}


int
_info(struct modinfo *modinfop)
{
    return mod_info(&audiov8233_modlinkage, modinfop);
}


/*
 * Allocate a list of DMA memory buffers (for playback or record)
 * and place the 8K chunks on the buffers free list.
 */
static int
audiov8233_alloc_sample_buffer(struct audiov8233_softstate *softstate,
			      sample_buffer_t *buf)
{
    int offs;
    struct sgd_chunk *chunk;

    for (offs = 0; offs < SAMPLEBUF_SIZE; offs += SAMPLEBUF_INTR_CHUNK) {

	chunk = kmem_zalloc(sizeof(struct sgd_chunk), KM_NOSLEEP);
	if (chunk == NULL)
	    return AUDIOV8233_FAILURE;

	chunk->next = buf->free;
	buf->free = chunk;

	chunk->dma_bound = FALSE;

	if (ddi_dma_alloc_handle(softstate->dip,
				 &dma_attr,
				 DDI_DMA_DONTWAIT, NULL,
				 &chunk->dma_handle) != DDI_SUCCESS) {
	    chunk->dma_handle = 0;
	    return AUDIOV8233_FAILURE;
	}
    
	if (ddi_dma_mem_alloc(chunk->dma_handle,
			      SAMPLEBUF_INTR_CHUNK,
			      &dev_attr,
			      DDI_DMA_STREAMING,
			      DDI_DMA_DONTWAIT, NULL,
			      &chunk->data, &chunk->reallen,
			      &chunk->acc_handle) != DDI_SUCCESS) {
	    DPRINTF1(AUDV8233M_ATTACH, "alloc dma sample buffer failed");
	    return AUDIOV8233_FAILURE;
	}
    }

    return AUDIOV8233_SUCCESS;
}


/*
 * Release the DMA memory sample buffers allocated with
 * audiov8233_alloc_sample_buffer()
 */
static int
audiov8233_free_sample_buffer(struct audiov8233_softstate *softstate,
			     sample_buffer_t *buf)
{
    struct sgd_chunk *chunk, *next;

    for (chunk = buf->free; chunk != NULL; chunk = next) {
	next = chunk->next;

	if (chunk->acc_handle) {
	    if (chunk->dma_bound) {
		ddi_dma_unbind_handle(chunk->dma_handle);
		chunk->dma_bound = FALSE;
	    }

	    ddi_dma_mem_free(&chunk->acc_handle);
	}

	if (chunk->dma_handle) {
	    ddi_dma_free_handle(&chunk->dma_handle);
	}
	
	kmem_free(chunk, sizeof(struct sgd_chunk));
    }

    return AUDIOV8233_SUCCESS;
}


/*
 * Configure the PCI device, so that the device is ready to 
 * accept I/O requests.  This includes enabling PCI busmaster,
 * and anabling access to the device's I/O registers.
 */
static int
audiov8233_pci_setup(struct audiov8233_softstate *softstate)
{
    dev_info_t *dip = softstate->dip;
    ddi_acc_handle_t conf_handle = softstate->pci_conf_handle;    
    uint16_t pci_status;
    uint16_t pci_vendorid;
    uint16_t pci_deviceid;
    uint8_t  pci_revision;
    uint8_t  pci_intrinfo;
    uint16_t pci_commandreg;
    char *revstr = "";

    pci_status   = pci_config_get16(conf_handle, PCI_CONF_STAT);
    pci_vendorid = pci_config_get16(conf_handle, PCI_CONF_VENID);
    pci_deviceid = pci_config_get16(conf_handle, PCI_CONF_DEVID);
    pci_revision = pci_config_get8 (conf_handle, PCI_CONF_REVID);
    pci_intrinfo = pci_config_get8 (conf_handle, PCI_CONF_ILINE);    
    
    /* try to find out a nice hardware name for the given pci hw */
    if (pci_vendorid == 0x1106) {
	/* via: */
	switch (pci_deviceid) {
	case 0x3059: revstr = pci_revision < 0x50 ? "VT8233" : "VT8235"; break;
	}
    }
    softstate->rev = pci_revision;
    softstate->revstr = revstr;

    /* Announce the hw we're attaching to */
    ddi_report_dev(dip);
    cmn_err(CE_CONT, "?PCI-device: %s, audiov8233%d, %s%s(rev. 0x%02x) at irq %d\n",
	    ddi_node_name(dip),
	    ddi_get_instance(dip), 
	    revstr, revstr[0] ? " " : "", 
	    pci_revision,
	    pci_intrinfo);

#if i386
    /*
     * reject interrupt line 255, per PCI spec 6.2.4, footnote 43,
     * this means "unknown" or "not connected" to the interrupt
     * controller on x86 based PCs.
     * Sun drivers (iprb) reject iline 0 as well.
     */
    if (pci_intrinfo == 0 || pci_intrinfo == 255) {
	cmn_err(CE_WARN, "interrupt line value out of range: %d", pci_intrinfo);
	return AUDIOV8233_FAILURE;
    }
#endif
    
    pci_commandreg = pci_config_get16(conf_handle, PCI_CONF_COMM);
    DPRINTF4(AUDV8233M_ATTACH, "pci command register 0x%x %s%s%s%s",
	    pci_commandreg,
	    pci_commandreg & PCI_COMM_SERR_ENABLE ? " serr#_en"  : "",
	    pci_commandreg & PCI_COMM_ME  ? " busmaster" : "",
	    pci_commandreg & PCI_COMM_MAE ? " mem_access" : "",
	    pci_commandreg & PCI_COMM_IO  ? " io_access" : "");
    if ((pci_commandreg & (PCI_COMM_IO|PCI_COMM_ME)) == (PCI_COMM_IO|PCI_COMM_ME)) {
	DPRINTF4(AUDV8233M_ATTACH, "busmastering and io_access already enabled");
    } else {
	DPRINTF4(AUDV8233M_ATTACH, "Enabling PCI I/O and bus mastering access");
	pci_config_put16(conf_handle, PCI_CONF_COMM, pci_commandreg | PCI_COMM_IO | PCI_COMM_ME);
    }

    return AUDIOV8233_SUCCESS;
}


static int
audiov8233_wait_for_codec_idle(struct audiov8233_softstate *softstate)
{
    int loop;
    uint32_t value = 0;

    for (loop = 100; --loop >= 0; ) {
        value = V8233_GET32(V8233_AC97_COMMAND);
        if ((value & V8233_AC97_CMD_CONTROLLER_BUSY) == 0) {
            /* codec is idle now */
            return AUDIOV8233_SUCCESS;
        }
        drv_usecwait(10);
    }

    DPRINTF1(AUDV8233M_ALL, "timeout, codec not idle [%x]", value);
    return AUDIOV8233_FAILURE;
}

static int
audiov8233_wait_for_codecdata(struct audiov8233_softstate *softstate, uint16_t *data)
{
    int loop;
    uint32_t value = 0;

    for (loop = 100; --loop >= 0; ) {
        value = V8233_GET32(V8233_AC97_COMMAND);
        if ((value & (V8233_AC97_CMD_PRIMARY_VALID|V8233_AC97_CMD_CONTROLLER_BUSY))
	    	== V8233_AC97_CMD_PRIMARY_VALID) {
           /* codec not busy any more + our requested data is ready */
            *data = value & V8233_AC97_CMD_DATA_MASK;
            return AUDIOV8233_SUCCESS;
        }
        drv_usecwait(10);
    }

    DPRINTF1(AUDV8233M_ALL, "timeout waiting got codec read [%x]", value);
    return AUDIOV8233_FAILURE;
}

/*
 * Read an AC97 primary codec mixer register
 *
 * Returns TRUE on success with the data read returned via the
 * *data pointer.  Returns FALSE in case the primary codec is not
 * ready any more for some reason.
 */
static int
audiov8233_ac97_read(void *closure, uint16_t sub_addr, uint16_t *data)
{
    struct audiov8233_softstate *softstate = closure;
    uint32_t codecreg = 0;
    uint16_t subreg;

    codecreg = V8233_AC97_CMD_PRIMARY_CODEC |
	V8233_AC97_CMD_PRIMARY_VALID |
	V8233_AC97_CMD_READ_ACCESS |
	V8233_AC97_CMD_REG_IDX(sub_addr);
    
    if (audiov8233_wait_for_codec_idle(softstate) != AUDIOV8233_SUCCESS) {
	return FALSE;
    }

    V8233_PUT32(V8233_AC97_COMMAND, codecreg);

    /* wait and read the codec register */
    if (audiov8233_wait_for_codecdata(softstate, &subreg) != AUDIOV8233_SUCCESS) {
        return FALSE;
    }
    *data = subreg;

    return TRUE;
}


/*
 * Write to an AC97 primary codec mixer register.
 *
 * Returns TRUE on success. Returns FALSE in
 * case the primary codec is not ready any more for some reason.
 */
static int
audiov8233_ac97_write(void *closure, uint16_t sub_addr, uint16_t value)
{
    struct audiov8233_softstate *softstate = closure;
    uint32_t codecvalue;

    codecvalue = V8233_AC97_CMD_REG_IDX(sub_addr) |
	(value & V8233_AC97_CMD_DATA_MASK);

    codecvalue |= V8233_AC97_CMD_PRIMARY_CODEC | V8233_AC97_CMD_WRITE_ACCESS;

    if (audiov8233_wait_for_codec_idle(softstate) != AUDIOV8233_SUCCESS) {
	return FALSE;
    }

    V8233_PUT32(V8233_AC97_COMMAND, codecvalue);

    return TRUE;
}


/*
 * Completely disable all VIA audio controller DMA
 */
static void
audiov8233_stop_dma(struct audiov8233_softstate *softstate)
{
    V8233_PUT8(V8233_RDCHN_REG(V8233_SGD_RDCHN_CTRL), V8233_SGD_CTRL_TERMINATE);
    V8233_PUT8(V8233_WRCHN_REG(V8233_SGD_WRCHN_CTRL), V8233_SGD_CTRL_TERMINATE);
    drv_usecwait(50);
    V8233_PUT8(V8233_RDCHN_REG(V8233_SGD_RDCHN_STAT), 
	       V8233_SGD_STAT_STOPPED|V8233_SGD_STAT_EOL|V8233_SGD_STAT_FLAG);
    V8233_PUT8(V8233_WRCHN_REG(V8233_SGD_WRCHN_STAT),
	       V8233_SGD_STAT_STOPPED|V8233_SGD_STAT_EOL|V8233_SGD_STAT_FLAG);
}


/*
 * reset the VIA audiocontroller's AC'97 link and reset the AC'97
 * codec to default values
 */
static int
audiov8233_reset_card(struct audiov8233_softstate *softstate,
		     audio_info_t		*audio_info,
		     int			 verbose)
{
    ddi_acc_handle_t conf_handle = softstate->pci_conf_handle;
    uint16_t codecval;
    uint8_t ac97_ctrl;
    uint8_t ac97_stat = 0;
    int loop;

    ac97_ctrl = pci_config_get8(conf_handle, PCI_CONF_V8233_AC97_CTRL);
    if (ac97_ctrl & PCI_V8233_ACCTL_ACRST) {
	/* reset is de-asserted, perform a warm AC97 reset:
	 * drive SYNC high for a minimum of 1us */
	pci_config_put8(conf_handle, PCI_CONF_V8233_AC97_CTRL, 
			PCI_V8233_ACCTL_ENAC97 |
			PCI_V8233_ACCTL_ACRST  | 
			PCI_V8233_ACCTL_RSYNC_HI);
	drv_usecwait(10);
	pci_config_put8(conf_handle, PCI_CONF_V8233_AC97_CTRL,
			PCI_V8233_ACCTL_ENAC97 |
			PCI_V8233_ACCTL_ACRST);
    } else {
	/* cold AC97 reset: 
	 * assert reset# low for the minimum specified time ... */
	pci_config_put8(conf_handle, PCI_CONF_V8233_AC97_CTRL, 0);
	drv_usecwait(100);
	/* ... then de-asserting reset# high */
	pci_config_put8(conf_handle, PCI_CONF_V8233_AC97_CTRL,
			PCI_V8233_ACCTL_ACRST);
    }

    pci_config_put8(conf_handle, PCI_CONF_V8233_AC97_CTRL,
		    PCI_V8233_ACCTL_ENAC97 |
		    PCI_V8233_ACCTL_ACRST | 
		    PCI_V8233_ACCTL_VSR_ON_DEMAND |
		    PCI_V8233_ACCTL_SGD_READ_CHANNEL);

    /* wait for ac97 codec ready status */
    for (loop = 6000; --loop >= 0; ) {
	ac97_stat = pci_config_get8(conf_handle, PCI_CONF_V8233_AC97_STAT);
	if (ac97_stat & PCI_V8233_ACSTAT_PRIMARY_CODEC_READY)
	    break;
	drv_usecwait(100);
    }
    if (loop < 0) {
	DPRINTF2(AUDV8233M_ATTACH,
		 "AC97 cold/warm reset still in progress? [%02x,%02x]\n",
		 ac97_ctrl, ac97_stat);
	return AUDIOV8233_FAILURE;
    }

    /* no DMA (not yet) */
    audiov8233_stop_dma(softstate);

    if (!CODEC_RESET(softstate->codec_handle, audio_info, verbose)) {
        DPRINTF2(AUDV8233M_CODEC, "Audio CODEC reset failed");
        return AUDIOV8233_FAILURE;
    }

    audiov8233_ac97_read (softstate, AC97_EXTENDED_AUDIO_REGISTER, &codecval);
    softstate->ac97_ext_features = codecval;

    return AUDIOV8233_SUCCESS;
}


/*
 * Allocate DMA resources for the VIA's SGD buffer descriptor lists.  We
 * allocate adjacent DMA memory in one physical page for all three DMA
 * engines available with the VIA audio controller.
 */
static int
audiov8233_setup_dma(struct audiov8233_softstate *softstate)
{
    int i, ret;
    uint_t ccount;
    ddi_dma_cookie_t cookie;

    softstate->sgd_dma_bound = FALSE;

    if ((ret = ddi_dma_alloc_handle(softstate->dip, &sgd_dma_attr,
				    DDI_DMA_DONTWAIT, NULL,
				    &softstate->sgd_dma_handle)) != DDI_SUCCESS) {
	DPRINTF4(AUDV8233M_ATTACH, "setup_dma: alloc handle failed (%d)\n", ret);
	softstate->sgd_dma_handle = 0;
	return AUDIOV8233_FAILURE;
    }

    if (ddi_dma_mem_alloc(softstate->sgd_dma_handle,
			  sizeof(v8233_sgd_t),
			  &dev_attr,
			  DDI_DMA_CONSISTENT,
			  DDI_DMA_DONTWAIT, NULL,
			  (caddr_t*) &softstate->sgd,
			  &softstate->sgdlen,
			  &softstate->sgd_acc_handle) != DDI_SUCCESS) {
	DPRINTF4(AUDV8233M_ATTACH, "setup_dma: mem alloc failed\n");
	return AUDIOV8233_FAILURE;
    }

    if ((ret = ddi_dma_addr_bind_handle(softstate->sgd_dma_handle,
					NULL,
					(caddr_t)softstate->sgd, softstate->sgdlen,
					DDI_DMA_RDWR|DDI_DMA_CONSISTENT,
					DDI_DMA_DONTWAIT, NULL,
					&cookie, &ccount)) != DDI_DMA_MAPPED) {
	DPRINTF4(AUDV8233M_ATTACH, "setup_dma: bind failed (%d)\n", ret);
	return AUDIOV8233_FAILURE;
    }

    /*
     * We should receive exactly one physical page mapping
     * (=contiguous physical memory)
     */
    if (ccount != 1) {
	DPRINTF1(AUDV8233M_ATTACH, "unexpected dma cookie count > 1 (%u)\n", ccount);
	return AUDIOV8233_FAILURE;
    }

    softstate->sgd_dma_bound = TRUE;

    DPRINTF4(AUDV8233M_ATTACH, "sgd at physical address %p\n",
	     (void*) (long) cookie.dmac_address);

    softstate->sgd_paddr = (v8233_sgd_t*) (long) cookie.dmac_address;

    bzero(softstate->sgd, sizeof(v8233_sgd_t));
    for (i = 0; i < V8233_NUM_SGD; i++) {
	softstate->sgd->pcm_out_sgd[i].bd_flag = TRUE;
	softstate->sgd->pcm_out_sgd[i].bd_stop = TRUE;

	softstate->sgd->pcm_in_sgd[i].bd_flag = TRUE;
	softstate->sgd->pcm_in_sgd[i].bd_stop = TRUE;
    }

    return AUDIOV8233_SUCCESS;
}


/*
 * Cleanup all resources allocated by the driver during attach time
 * and during run-time, in reverse order.
 */
static int
audiov8233_cleanup(struct audiov8233_softstate *softstate)
{
    if (softstate->audio_handle) {
	if (audio_hlp_detach(softstate->audio_handle, DDI_DETACH) != DDI_SUCCESS)
	    return DDI_FAILURE;
    }

    if (softstate->codec_handle) {
        CODEC_DELETE(softstate->codec_handle);
        softstate->codec_handle = NULL;
    }

    if (softstate->sgd_acc_handle != NULL) {

	/*
	 * just to be sure, turn off DMA, before releaseing the DMA
	 * sample buffers
	 */
	audiov8233_stop_dma(softstate);

	/* free audio sample buffer DMA memory */
	audiov8233_free_sample_buffer(softstate, &softstate->play_buf);
	audiov8233_free_sample_buffer(softstate, &softstate->record_buf);

	/* free SGD DMA memory for the three VIA DMA engines */
	if (softstate->sgd_dma_bound)
	    ddi_dma_unbind_handle(softstate->sgd_dma_handle);
	ddi_dma_mem_free(&softstate->sgd_acc_handle);
	softstate->sgd = NULL;
    }

    if (softstate->sgd_dma_handle) {
	ddi_dma_free_handle(&softstate->sgd_dma_handle);
    }

    /*
     * remove interrupt handler before we free the card's mapped registers
     * otherwise an interrupt could access softstate->regs_handle in our
     * interrupt handler
     */
    if (softstate->intr_cookie != NULL) {
	ddi_remove_intr(softstate->dip, 0, softstate->intr_cookie);
    }

    if (softstate->regs_handle != NULL) {
	ddi_regs_map_free(&softstate->regs_handle);
    }

    if (softstate->pci_conf_handle != NULL) {
	pci_config_teardown(&softstate->pci_conf_handle);
    }

    mutex_destroy(&softstate->instance_mutex);
    mutex_destroy(&softstate->intr_mutex);

    ddi_soft_state_free(audiov8233_state, ddi_get_instance(softstate->dip));

    return DDI_SUCCESS;
}


/*
 * Playback of a chunk of sample data has finished, release the DMA buffer
 * to the free list.
 */
static void
audiov8233_release_play_buf_head(struct audiov8233_softstate *softstate,
				sample_buffer_t *buf)
{
    struct sgd_chunk *chunk;
    v8233_buf_descr_t *bd;

    bd = softstate->sgd->pcm_out_sgd + buf->head;
    bd->bd_len  = 0;
    bd->bd_stop = TRUE;
    bd->bd_flag = TRUE;

    if ((chunk = buf->active[buf->head]) != NULL) {
	DPRINTF4(AUDV8233M_INTR, "free chunk at index %d\n", buf->head);
	if (chunk->dma_bound) {
	    ddi_dma_unbind_handle(chunk->dma_handle);
	    chunk->dma_bound = FALSE;
	}
	chunk->next = buf->free;
	buf->free = chunk;
	buf->active[buf->head] = NULL;
    }
    ++buf->buffers_available;
    if (++buf->head >= V8233_NUM_SGD)
	buf->head = 0;
}


/*
 * Recoding to a chunk of sample data at the front of the DMA buffer
 * descriptor list has finished.  Send the recorded audio samples to
 * the audio driver framework and release the buffer to the free list.
 */
static void
audiov8233_release_record_buf_head(struct audiov8233_softstate *softstate,
				  sample_buffer_t *buf)
{
    struct sgd_chunk *chunk;
    int byte_to_sample_shift;

    byte_to_sample_shift = 0;
    if (softstate->audio_info.record.channels == 2) byte_to_sample_shift++;
    if (softstate->audio_info.record.precision == 16) byte_to_sample_shift++;

    if ((chunk = buf->active[buf->head]) != NULL) {
	DPRINTF4(AUDV8233M_INTR, "free chunk at index %d, %d samples\n",
		 buf->head,
		 SAMPLEBUF_INTR_CHUNK >> byte_to_sample_shift);
	if (chunk->dma_bound) {
	    ddi_dma_unbind_handle(chunk->dma_handle);
	    chunk->dma_bound = FALSE;
	}

	audio_hlp_send_audio(softstate->audio_handle, chunk->data, 0,
			     SAMPLEBUF_INTR_CHUNK >> byte_to_sample_shift);

	chunk->next = buf->free;
	buf->free = chunk;
	buf->active[buf->head] = NULL;
    }
    ++buf->buffers_available;
    if (++buf->head >= V8233_NUM_SGD)
	buf->head = 0;
}


/*
 * interrupt handler for the audiov8233 device.
 */
static uint_t
audiov8233_intr(caddr_t intr_arg)
{
    struct audiov8233_softstate *softstate = (void*)intr_arg;
    sample_buffer_t *buf;
    uint8_t play_stat, rec_stat;
    /*
    uint32_t cur_ptr;
    uint32_t cur_cnt;
    */
    int cur_idx;
    int claimed = DDI_INTR_UNCLAIMED;

    DPRINTF4(AUDV8233M_INTR, "audiov8233_intr\n");

    play_stat = V8233_GET8(V8233_RDCHN_REG(V8233_SGD_RDCHN_STAT));
    if (play_stat & (V8233_SGD_STAT_STOPPED|V8233_SGD_STAT_FLAG|V8233_SGD_STAT_EOL)) {
	V8233_PUT8(V8233_RDCHN_REG(V8233_SGD_RDCHN_STAT),
		   play_stat & (V8233_SGD_STAT_STOPPED|V8233_SGD_STAT_FLAG|V8233_SGD_STAT_EOL));
	claimed = DDI_INTR_CLAIMED;

	cur_idx = V8233_GET8(V8233_RDCHN_REG(V8233_SGD_RDCHN_CUR_IDX));
	/*
	cur_cnt = V8233_GET32(V8233_RDCHN_REG(V8233_SGD_RDCHN_CUR_CNT));
	cur_ptr = V8233_GET32(V8233_RDCHN_REG(V8233_SGD_RDCHN_PTR));
	cur_idx = (cur_ptr - (uint32_t)softstate->sgd_paddr->pcm_out_sgd)
	    / sizeof(v8233_buf_descr_t);
	*/

	DPRINTF4(AUDV8233M_INTR, "PCM out irq, stat 0x%x, idx %d\n",
		 play_stat, cur_idx);

	buf = &softstate->play_buf;

	/* free buffers at the head of the buffer desc list */
	mutex_enter(&softstate->intr_mutex);
	while (buf->head != cur_idx)
	    audiov8233_release_play_buf_head(softstate, buf);
	mutex_exit(&softstate->intr_mutex);

	/* prepare new playback buffers and append them at the tail of the list */
	DPRINTF4(AUDV8233M_INTR, "fill more play buffers\n");
	if (audiov8233_fill_play_buffers(softstate) == 0 &&
	    (play_stat & V8233_SGD_STAT_STOPPED)) {

	    DPRINTF4(AUDV8233M_INTR,
		     "out of data, shutdown playback, status %02x\n",
		     play_stat);
	    audio_hlp_play_shutdown(softstate->audio_handle, 0);
	    
	    mutex_enter(&softstate->intr_mutex);
	    V8233_PUT8(V8233_RDCHN_REG(V8233_SGD_RDCHN_CTRL), V8233_SGD_CTRL_TERMINATE);
	    softstate->play_buf.io_started = FALSE;
	    audiov8233_release_play_buf_head(softstate, buf);
	    mutex_exit(&softstate->intr_mutex);

	    if (softstate->play_timeout_id) {
		untimeout(softstate->play_timeout_id);
		softstate->play_timeout_id = 0;
	    }
	}
    }

    rec_stat  = V8233_GET8(V8233_WRCHN_REG(V8233_SGD_WRCHN_STAT));
    if (rec_stat & (V8233_SGD_STAT_STOPPED|V8233_SGD_STAT_FLAG|V8233_SGD_STAT_EOL)) {
	V8233_PUT8(V8233_WRCHN_REG(V8233_SGD_WRCHN_STAT),
		   rec_stat & (V8233_SGD_STAT_STOPPED|V8233_SGD_STAT_FLAG|V8233_SGD_STAT_EOL));
	claimed = DDI_INTR_CLAIMED;

	cur_idx = V8233_GET8(V8233_WRCHN_REG(V8233_SGD_WRCHN_CUR_IDX));
	/*
	cur_ptr = V8233_GET32(V8233_WRCHN_REG(V8233_SGD_WRCHN_PTR));
	cur_idx = (cur_ptr - (uint32_t)softstate->sgd_paddr->pcm_in_sgd)
	    / sizeof(v8233_buf_descr_t);
	*/

	DPRINTF4(AUDV8233M_INTR, "PCM in irq, stat 0x%x, cur_idx 0x%02x\n",
		 rec_stat, cur_idx);


	buf = &softstate->record_buf;

	/*
	 * send completed record buffers from the head of the descriptor list
	 * to the audio framework and release the buffers
	 */
	mutex_enter(&softstate->intr_mutex);
	while (buf->head != cur_idx)
	    audiov8233_release_record_buf_head(softstate, buf);
	mutex_exit(&softstate->intr_mutex);

	/* prepare new record buffers and append them at the tail of the list */
	audiov8233_prepare_record_buffers(softstate);
    }

    return claimed;
}


static audio_device_t audiov8233_dev_info;
static ah_ad_entry_t audiov8233_entry;
static ah_ad_ch_cap_t audiov8233_ch_cap;
static ah_ad_cap_comb_t audiov8233_cap_comb[];
static ah_ad_sample_rates_t audiov8233_sample_rates_48k;

static int
audiov8233_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
    int instance = ddi_get_instance(dip);
    struct audiov8233_softstate *softstate;
    int status;
    char *getdev_name;

    DPRINTF4(AUDV8233M_ATTACH, "attach, cmd=%d\n", cmd);

    switch (cmd) {
    case DDI_ATTACH:
	break;

    case DDI_RESUME:
    case DDI_PM_RESUME:
    default:
	return DDI_FAILURE;
    }

    /*
     * this card is a bus master, reject any slave-only slot
     */
    if (ddi_slaveonly(dip) == DDI_SUCCESS) {
        cmn_err(CE_NOTE, "slot does not support PCI bus-master");
        return DDI_FAILURE;
    }

    /*
     * Make sure we're not using a high level interrupt handler.
     */
    if (ddi_intr_hilevel(dip, 0)) {
	cmn_err(CE_NOTE, "Driver does not support high level interrupt.");
	cmn_err(CE_NOTE, "You should probably add 'interrupt-priorities=9;' to audiov8233.conf");
	return DDI_FAILURE;
    }

    if (ddi_soft_state_zalloc(audiov8233_state, instance) != DDI_SUCCESS)
	return DDI_FAILURE;

    softstate = ddi_get_soft_state(audiov8233_state, instance);
    softstate->dip = dip;

    if (pci_config_setup(softstate->dip, &softstate->pci_conf_handle) != DDI_SUCCESS) {
	DPRINTF4(AUDV8233M_ATTACH, "audiov8233%d: pci_config_setup failed", instance);
	audiov8233_cleanup(softstate);
	return DDI_FAILURE;
    }

    if (audiov8233_pci_setup(softstate) != AUDIOV8233_SUCCESS) {
	DPRINTF4(AUDV8233M_ATTACH, "audiov8233%d: audiov8233_pci_setup failed", instance);
	audiov8233_cleanup(softstate);
	return DDI_FAILURE;	
    }

#if	USE_STANDARD_PLAYBACK_CHANNELS
    /*
     * Use ordinary playback channel, not available on all revs. of VIA8233 chipset.
     */
    softstate->playback_reg_base = V8233_SGD_RDCHN_IOBASE;
#else
    /*
     * multichannel playback is available on all VIA8233 chipset revs.
     */
    softstate->playback_reg_base = V8233_SGD_MP_IOBASE;
#endif

    bcopy(&audiov8233_defaults, &softstate->audio_info, sizeof(audio_info_t));
    bcopy(&audiov8233_dev_info, &softstate->dev_info, sizeof(audio_device_t));

    /* allow spoofing of audio_device.name in AUDIO_GETDEV ioctl */
    if (ddi_prop_lookup_string(DDI_DEV_T_ANY, dip,
			       DDI_PROP_DONTPASS | DDI_PROP_NOTPROM,
			       "ioctl_audio_getdev_name",
			       &getdev_name) == DDI_PROP_SUCCESS) {

	strncpy(softstate->dev_info.name, getdev_name,
		sizeof(softstate->dev_info.name));
	ddi_prop_free(getdev_name);
    }

    sprintf(softstate->dev_info.version, "%02X,%s",
	    softstate->rev, softstate->revstr);

    softstate->ad_info.ad_int_vers = AD_VERS1;
    softstate->ad_info.ad_defaults = &audiov8233_defaults;
    bcopy(&audiov8233_ch_cap, &softstate->ad_info.ad_play,
	  sizeof(ah_ad_ch_cap_t));
    bcopy(&audiov8233_ch_cap, &softstate->ad_info.ad_record,
	  sizeof(ah_ad_ch_cap_t));
    softstate->ad_info.ad_play_comb = audiov8233_cap_comb;
    softstate->ad_info.ad_rec_comb = audiov8233_cap_comb;
    softstate->ad_info.ad_entry    = &audiov8233_entry;
    softstate->ad_info.ad_dev_info = &softstate->dev_info;
    softstate->ad_info.ad_translate_flags =
				AD_MISC_8_P_TRANSLATE | AD_MISC_8_R_TRANSLATE;

    softstate->audio_handle = audio_hlp_attach(dip, cmd, &softstate->ad_info);
    if (!softstate->audio_handle) {
	cmn_err(CE_WARN, "failed to attach audio helper module");
	audiov8233_cleanup(softstate);
	return DDI_FAILURE;
    }

    audio_hlp_set_private(softstate->audio_handle, softstate);


    /* map the audio controller's register block */
    status = ddi_regs_map_setup(dip, 1,
				(caddr_t *)&softstate->regs,
				0, V8233_REGS_SIZE,
				&dev_attr, &softstate->regs_handle);

    if (status != DDI_SUCCESS) {
	if (status == DDI_REGS_ACC_CONFLICT) {
	    cmn_err(CE_WARN, "audio controller regs mapping conflicted with other mapping");
	} else {
	    cmn_err(CE_WARN, "audio controller regs mapping failed");
	}
	audiov8233_cleanup(softstate);
	return DDI_FAILURE;
    }


    mutex_init(&softstate->instance_mutex, NULL, MUTEX_DRIVER, NULL);

    if (ddi_get_iblock_cookie(dip, 0, &softstate->intr_cookie) != DDI_SUCCESS) {
	cmn_err(CE_WARN, "Could not get iblock cookie");
	audiov8233_cleanup(softstate);
	return DDI_FAILURE;
    }
    mutex_init(&softstate->intr_mutex, NULL,
	       MUTEX_DRIVER, (void*)softstate->intr_cookie);

    if (ddi_add_intr(dip, 0,
		     &softstate->intr_cookie, NULL,
		     audiov8233_intr, (caddr_t)softstate) != DDI_SUCCESS) {

	cmn_err(CE_WARN, "Could not add interrupt handler");
	audiov8233_cleanup(softstate);
	return DDI_FAILURE;
    }

    /*
     * try to reset the audio controller, this verifies basic operation of the
     * sound hardware.
     */
    
    {
	static struct ac97_codec_ops codec_ops = {
	    audiov8233_ac97_read,
	    audiov8233_ac97_write
	};

	softstate->codec_handle = ac97_codec_init(softstate->audio_handle, &codec_ops, softstate);
    }

    if (audiov8233_reset_card(softstate, &audiov8233_defaults, TRUE) != AUDIOV8233_SUCCESS) {
	cmn_err(CE_WARN, "reset of v8233 audio controller failed");
	audiov8233_cleanup(softstate);
	return DDI_FAILURE;	
    }


    if (audiov8233_setup_dma(softstate) != AUDIOV8233_SUCCESS) {
	cmn_err(CE_WARN, "dma setup of v8233 audio controller failed");
	audiov8233_cleanup(softstate);
	return DDI_FAILURE;	
    }

    if (audiov8233_alloc_sample_buffer(softstate, &softstate->play_buf) != AUDIOV8233_SUCCESS
	|| audiov8233_alloc_sample_buffer(softstate, &softstate->record_buf) != AUDIOV8233_SUCCESS) {
	cmn_err(CE_WARN, "cannot allocate sample buffers");
	audiov8233_cleanup(softstate);
	return DDI_FAILURE;
    }

    /*
     * if we have an AC'97 codec without sample rate converter, 48kHz is all we can
     * play/record
     */
    if (softstate->playback_reg_base != V8233_SGD_RDCHN_IOBASE &&
	!(softstate->ac97_ext_features & EAR_VRA)) {
	bcopy(&audiov8233_sample_rates_48k, &softstate->ad_info.ad_play.ad_srs,
	      sizeof(ah_ad_sample_rates_t));
	bcopy(&audiov8233_sample_rates_48k, &softstate->ad_info.ad_record.ad_srs,
	      sizeof(ah_ad_sample_rates_t));
    }

    return DDI_SUCCESS;    
}


static int
audiov8233_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
    int instance = ddi_get_instance(dip);
    struct audiov8233_softstate *softstate = ddi_get_soft_state(audiov8233_state, instance);

    DPRINTF4(AUDV8233M_ATTACH, "detach, cmd=%d\n", cmd);

    switch (cmd) {
    case DDI_DETACH:
	break;

    default:
	return DDI_FAILURE;
    }

    if (audiov8233_cleanup(softstate) != DDI_SUCCESS) 
	return DDI_FAILURE;

    return DDI_SUCCESS;
}


/*
 * Audio Driver functions (for ah_ad_entry_t)
 */

static int    
audiov8233_setup(audiohdl_t handle, int stream, int flags)
{
    struct audiov8233_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUDV8233M_DRIVER, "audiov8233_setup, stream %d, flags %x\n", stream, flags);

    softstate->play_buf.io_started = FALSE;
    softstate->record_buf.io_started = FALSE;

    return AUDIO_SUCCESS;
}


static void
audiov8233_teardown(audiohdl_t handle, int stream, int dir)
{
    DPRINTF4(AUDV8233M_DRIVER, "audiov8233_teardown, stream %d, dir %d\n", stream, dir);
}


static int
audiov8233_set_config(audiohdl_t handle, int stream,
		     int command, int dir, int arg1, int arg2)
{
    struct audiov8233_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUDV8233M_DRIVER, "audiov8233_set_config, stream %d, "
	     "command %d, dir %d, arg1 %d, arg2 %d\n",
	     stream, command, dir, arg1, arg2);

    CODEC_SET_CONFIG(softstate->codec_handle, command, dir, arg1, arg2);

    return AUDIO_SUCCESS;
}


static int
audiov8233_set_format(audiohdl_t handle, int stream,
		 int dir, int sample_rate, int channels,
		 int precision, int encoding)
{
    struct audiov8233_softstate *softstate = audio_hlp_get_private(handle);
    uint32_t type;
    uint32_t slot;

    DPRINTF4(AUDV8233M_DRIVER, "audiov8233_set_format, stream %d, dir %d, "
	     "sample_rate %d, channels %d, precision %d, encoding %d\n",
	     stream, dir, sample_rate, channels, precision, encoding);

    if (dir == AUDIO_PLAY &&
	softstate->playback_reg_base != V8233_SGD_RDCHN_IOBASE &&
	!(softstate->ac97_ext_features & EAR_VRA) && 
	sample_rate != AC97_SAMPLE_RATE_48000) {
	/* no variable sample rates in multichannel mode? */
	DPRINTF2(AUDV8233M_DRIVER, "unsupported sample rate %d\n", sample_rate);
	return AUDIO_FAILURE;
    }

    if (dir == AUDIO_RECORD &&
	!(softstate->ac97_ext_features & EAR_VRA) && 
	sample_rate != AC97_SAMPLE_RATE_48000) {
	DPRINTF2(AUDV8233M_DRIVER, "unsupported sample rate %d\n", sample_rate);
	return AUDIO_FAILURE;
    }

    mutex_enter(&softstate->intr_mutex);

    audiov8233_ac97_write(softstate, AC97_EXTENDED_AUDIO_STAT_CTRL_REGISTER, EASCR_VRA);
    if (dir == AUDIO_PLAY) {
	uint16_t val;

	/*
	 * we use the VIA audio controller SRC, so the AC97 codec
	 * stays at 48kHz (The audio controller SRC seems to work
	 * much better than the SRC included in the AC97 SRC on my
	 * spacewalker shuttle board - it produces no noise).
	 *
	 * In multichannel playback mode, we must use the AC97 codec
	 * SRC.
	 */
	DPRINTF4(AUDV8233M_DRIVER, "set DAC VRA %d\n", sample_rate);
	audiov8233_ac97_write(softstate, AC97_EXTENDED_FRONT_DAC_RATE_REGISTER,
			      softstate->playback_reg_base == V8233_SGD_RDCHN_IOBASE
			      		? 48000
			      		: sample_rate);
	
	audiov8233_ac97_read (softstate, AC97_EXTENDED_FRONT_DAC_RATE_REGISTER, &val);
	DPRINTF4(AUDV8233M_DRIVER, "Front DAC samplerate is %d\n", val);
    } else {
#ifndef	AC97_EXTENDED_LR_ADC_RATE_REGISTER
/* workaround a bug in sun's ac97.h header file: */
#define	AC97_EXTENDED_LR_ADC_RATE_REGISTER	AC97_EXTENDED_LR_DAC_RATE_REGISTER
#endif
	DPRINTF4(AUDV8233M_DRIVER, "set ADC VRA %d\n", sample_rate);
	audiov8233_ac97_write(softstate, AC97_EXTENDED_LR_ADC_RATE_REGISTER, sample_rate);
	audiov8233_ac97_write(softstate, AC97_EXTENDED_MIC_ADC_RATE_REGISTER, sample_rate);
	
    }

    if (dir == AUDIO_PLAY) {
	if (softstate->playback_reg_base == V8233_SGD_RDCHN_IOBASE) {
	    /* setup format for standard playback channels */
	    V8233_PUT8(V8233_RDCHN_REG(V8233_SGD_RDCHN_VOLL), 0);
	    V8233_PUT8(V8233_RDCHN_REG(V8233_SGD_RDCHN_VOLR), 0);

	    type = V8233_GET32(V8233_RDCHN_REG(V8233_SGD_RDCHN_TYPE));

	    if (precision == 16) 
		type |= V8233_SGD_TYPE_16BIT;
	    else
		type &= ~V8233_SGD_TYPE_16BIT;

	    if (channels == 2) 
		type |= V8233_SGD_TYPE_STEREO;
	    else
		type &= ~V8233_SGD_TYPE_STEREO;

	    type &= ~V8233_SGD_TYPE_VRA_FREQ_MASK;
	    type |= (0xffffu * (uint32_t)sample_rate / (48000/16)) & V8233_SGD_TYPE_VRA_FREQ_MASK;

	    V8233_PUT32(V8233_RDCHN_REG(V8233_SGD_RDCHN_TYPE), type);
	} else {
	    /* setup format for multichannel playback */
	    slot = V8233_GET32(V8233_RDCHN_REG(V8233_SGD_MP_SLOT_SELECT));
	    slot &= ~V8233_SGD_MP_SLOT_MASK;

	    if (precision == 16) 
		type = V8233_SGD_MP_TYPE_16BIT;
	    else
		type = V8233_SGD_MP_TYPE_8BIT;
	    
	    if (channels == 1) {
		type |= V8233_SGD_MP_TYPE_CHANNELS(1);
		slot |= V8233_SGD_MP_SLOT3(1) | V8233_SGD_MP_SLOT4(1);
	    } else {
		type |= V8233_SGD_MP_TYPE_CHANNELS(2);
		slot |= V8233_SGD_MP_SLOT3(1) | V8233_SGD_MP_SLOT4(2);
	    }
	    V8233_PUT8 (V8233_RDCHN_REG(V8233_SGD_MP_TYPE), type);
	    V8233_PUT32(V8233_RDCHN_REG(V8233_SGD_MP_SLOT_SELECT), slot);
	}

	softstate->audio_info.play.sample_rate = sample_rate;
	softstate->audio_info.play.channels    = channels;
	softstate->audio_info.play.precision   = precision;
	softstate->audio_info.play.encoding    = encoding;
    } else {
	V8233_PUT8(V8233_WRCHN_REG(V8233_SGD_WRCHN_FIFO), V8233_SGD_FIFO_ENABLE);

	type = V8233_GET32(V8233_WRCHN_REG(V8233_SGD_WRCHN_TYPE));

	if (precision == 16) 
	    type |= V8233_SGD_TYPE_16BIT;
	else
	    type &= ~V8233_SGD_TYPE_16BIT;

	if (channels == 2) 
	    type |= V8233_SGD_TYPE_STEREO;
	else
	    type &= ~V8233_SGD_TYPE_STEREO;

#if 0
	/* it seems there's no SRC for the recording side in the via audio controller...*/
	type &= ~V8233_SGD_TYPE_VRA_FREQ_MASK;
	type |= (0xffffu * 48000 / ((uint32_t)sample_rate/16)) & V8233_SGD_TYPE_VRA_FREQ_MASK;
#endif

	V8233_PUT32(V8233_WRCHN_REG(V8233_SGD_WRCHN_TYPE), type);
	/*printf("record chan type=0x%08x", V8233_GET32(V8233_WRCHN_REG(V8233_SGD_WRCHN_TYPE)));*/

	softstate->audio_info.record.sample_rate = sample_rate;
	softstate->audio_info.record.channels    = channels;
	softstate->audio_info.record.precision   = precision;
	softstate->audio_info.record.encoding    = encoding;
    }

    mutex_exit(&softstate->intr_mutex);

    return AUDIO_SUCCESS;
}


static void
audiov8233_timed_fill_play_buf(void *arg)
{
    struct audiov8233_softstate *softstate = arg;

    DPRINTF4(AUDV8233M_DRIVER, "audiov8233_timed_fill_play_buf\n");

    mutex_enter(&softstate->intr_mutex);
    softstate->play_timeout_id = 0;
    mutex_exit(&softstate->intr_mutex);

    audiov8233_fill_play_buffers(softstate);    
}


/*
 * Fetch free sample buffers from the free list, fill them with audio
 * samples abtained from the audio framework and add them at the tail
 * of the VIA's SGD.
 */
static int
audiov8233_fill_play_buffers(struct audiov8233_softstate *softstate)
{
    struct sgd_chunk *chunk;
    struct sample_buffer *buf = &softstate->play_buf;
    int bytes, len;
    int total_samples;
    int samples;
    int bindstat;
    int i;
    uint_t ccount;
    ddi_dma_cookie_t cookie;
    v8233_buf_descr_t *bdesc;
    uint8_t ctrl;
    int byte_to_sample_shift;

    total_samples = 0;

    byte_to_sample_shift = 0;
    if (softstate->audio_info.play.channels == 2) byte_to_sample_shift++;
    if (softstate->audio_info.play.precision == 16) byte_to_sample_shift++;

    mutex_enter(&softstate->intr_mutex);

    if (!buf->io_started) {

	V8233_PUT32 (V8233_RDCHN_REG(V8233_SGD_RDCHN_PTR), (uint32_t) (long) softstate->sgd_paddr->pcm_out_sgd);

	buf->head = 0;
	buf->tail = 0;
	buf->buffers_available = V8233_NUM_SGD;
    }

    while ((chunk = buf->free) != NULL) {
	buf->free = chunk->next;
	chunk->next = NULL;

	bytes = SAMPLEBUF_INTR_CHUNK;
	samples = bytes >> byte_to_sample_shift;

	samples = audio_hlp_get_audio(softstate->audio_handle, chunk->data, 0, samples);
	DPRINTF4(AUDV8233M_INTR, "got %d samples audio data\n", samples);
	if (samples <= 0) {
	    chunk->next = buf->free;
	    buf->free = chunk;
	    break;
	}
	bytes = samples << byte_to_sample_shift;
	total_samples += samples;

	bindstat = ddi_dma_addr_bind_handle(chunk->dma_handle,
					    NULL,
					    chunk->data, chunk->reallen,
					    DDI_DMA_WRITE,
					    DDI_DMA_DONTWAIT, NULL,
					    &cookie, &ccount);
	if (bindstat != DDI_DMA_MAPPED) {
	    DPRINTF4(AUDV8233M_INTR, "dma_addr_bind_handle failed %d\n", bindstat);
	    chunk->next = buf->free;
	    buf->free = chunk;
	    break;
	}

	chunk->dma_bound = TRUE;
	buf->active[buf->tail] = chunk;

	for (i = 0; i < ccount && bytes > 0; i++) {
	    if (i > 0)
		ddi_dma_nextcookie(chunk->dma_handle, &cookie);
	    DPRINTF4(AUDV8233M_INTR, "cookie #%d: at %p, %ld bytes\n",
		     i, (void*) (long) cookie.dmac_address, (long)cookie.dmac_size);

	    if (cookie.dmac_size < bytes)
		len = cookie.dmac_size;
	    else
		len = bytes;
	    bytes -= len;

	    buf->buffers_available--;

	    /*
	     * since we map DMA memory with the DDI_NEVERSWAP_ACC attribute, 
	     * we can access it without using ddi_{get,put}N
	     */
	    bdesc = softstate->sgd->pcm_out_sgd + buf->tail;
	    bdesc->bd_base = cookie.dmac_address;
	    bdesc->bd_len  = len;
	    bdesc->bd_stop = FALSE;
	    bdesc->bd_flag = (bytes <= 0);
	    bdesc->bd_eol  = buf->tail == V8233_NUM_SGD-1;

	    V8233_PUT8(V8233_RDCHN_REG(V8233_SGD_RDCHN_STOP_IDX), buf->tail);

	    if (++buf->tail >= V8233_NUM_SGD)
		buf->tail = 0;
	}
    }

    /* release pause status, start SGD */
    ctrl = V8233_GET8(V8233_RDCHN_REG(V8233_SGD_RDCHN_CTRL));
    ctrl &= ~V8233_SGD_CTRL_PAUSE;
    if (!buf->io_started && buf->head != buf->tail) {
	ctrl |= ( V8233_SGD_CTRL_START |
		  V8233_SGD_CTRL_AUTOSTART |
		  V8233_SGD_CTRL_INTR_STOP |
		  V8233_SGD_CTRL_INTR_EOL |
		  V8233_SGD_CTRL_INTR_FLAG );
	buf->io_started = TRUE;
    }
    V8233_PUT8(V8233_RDCHN_REG(V8233_SGD_RDCHN_CTRL), ctrl);

    if (buf->free && !softstate->play_timeout_id) {
	/*
	 * free sample buffers left, and no more sample data available from
	 * the framework. Schedule a timeout routine to retry filling more
	 * buffers in time before the VIA AC has used all the samples we added
	 * to the SGD.
	 * 
	 * XXX The timeout (20ms) could be tuned a bit.  At 48kHz, 16bit, stereo,
	 * an 8K sample buffer plays ~41ms. At lower sample rates the timeout
	 * could reduced.
	 */
	softstate->play_timeout_id = 
	    timeout(audiov8233_timed_fill_play_buf, softstate,
		    drv_usectohz(20000));
    }

    mutex_exit(&softstate->intr_mutex);

    DPRINTF4(AUDV8233M_INTR, "DMA enabled, PCM out, status 0x%02x, ctrl 0x%02x, ptr 0x%08x, type 0x%08x, cnt 0x%08x\n",
	     V8233_GET8(V8233_RDCHN_REG(V8233_SGD_RDCHN_STAT)),
	     V8233_GET8(V8233_RDCHN_REG(V8233_SGD_RDCHN_CTRL)),
	     V8233_GET32(V8233_RDCHN_REG(V8233_SGD_RDCHN_PTR)),
	     V8233_GET32(V8233_RDCHN_REG(V8233_SGD_RDCHN_TYPE)),
	     V8233_GET32(V8233_RDCHN_REG(V8233_SGD_RDCHN_CUR_CNT)));

    return total_samples;
}


/*
 * Fetch free sample buffers from the free list,and add them at the
 * tail of the ICH's BDL for recording audio samples.
 */
static int
audiov8233_prepare_record_buffers(struct audiov8233_softstate *softstate)
{
    struct sgd_chunk *chunk;
    struct sample_buffer *buf = &softstate->record_buf;
    int bindstat;
    int i;
    uint_t ccount;
    ddi_dma_cookie_t cookie;
    v8233_buf_descr_t *bdesc;
    uint8_t ctrl;

    mutex_enter(&softstate->intr_mutex);

    if (!buf->io_started) {
	V8233_PUT32(V8233_WRCHN_REG(V8233_SGD_WRCHN_PTR), (uint32_t) (long) softstate->sgd_paddr->pcm_in_sgd);

	buf->head = 0;
	buf->tail = 0;
	buf->buffers_available = V8233_NUM_SGD;
    }
    
    while ((chunk = buf->free) != NULL) {
	buf->free = chunk->next;
	chunk->next = NULL;

	bindstat = ddi_dma_addr_bind_handle(chunk->dma_handle,
					    NULL,
					    chunk->data, chunk->reallen,
					    DDI_DMA_READ,
					    DDI_DMA_DONTWAIT, NULL,
					    &cookie, &ccount);
	if (bindstat != DDI_DMA_MAPPED) {
	    DPRINTF4(AUDV8233M_INTR, "dma_addr_bind_handle failed %d\n", bindstat);
	    chunk->next = buf->free;
	    buf->free = chunk;
	    break;
	}

	chunk->dma_bound = TRUE;
	buf->active[buf->tail] = chunk;

	for (i = 0; i < ccount; i++) {
	    if (i > 0)
		ddi_dma_nextcookie(chunk->dma_handle, &cookie);
	    DPRINTF4(AUDV8233M_INTR, "cookie #%d: at %p, %ld bytes\n",
		     i, (void*) (long) cookie.dmac_address, (long)cookie.dmac_size);

	    buf->buffers_available--;

	    /*
	     * since we map DMA memory with the DDI_NEVERSWAP_ACC attribute, 
	     * we can access it without using ddi_{get,put}N
	     */
	    bdesc = softstate->sgd->pcm_in_sgd + buf->tail;
	    bdesc->bd_base = cookie.dmac_address;
	    bdesc->bd_len  = cookie.dmac_size;
	    bdesc->bd_stop = FALSE;
	    bdesc->bd_flag = i == ccount-1;
	    bdesc->bd_eol  = buf->tail == V8233_NUM_SGD-1;

	    V8233_PUT8(V8233_WRCHN_REG(V8233_SGD_WRCHN_STOP_IDX), buf->tail);
	    DPRINTF4(AUDV8233M_INTR, "entry %d prepared\n", buf->tail);
	    if (++buf->tail >= V8233_NUM_SGD)
		buf->tail = 0;
	}
    }

    if (!buf->io_started) {

	ctrl = V8233_GET8(V8233_WRCHN_REG(V8233_SGD_WRCHN_CTRL));
	ctrl |= ( V8233_SGD_CTRL_START |
		  V8233_SGD_CTRL_AUTOSTART |
		  V8233_SGD_CTRL_INTR_STOP |
		  V8233_SGD_CTRL_INTR_EOL |
		  V8233_SGD_CTRL_INTR_FLAG );
	V8233_PUT8(V8233_WRCHN_REG(V8233_SGD_WRCHN_CTRL), ctrl);

	buf->io_started = TRUE;
    }

    mutex_exit(&softstate->intr_mutex);

    return AUDIOV8233_SUCCESS;
}


static int
audiov8233_start_play(audiohdl_t handle, int stream)
{
    struct audiov8233_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUDV8233M_DRIVER, "audiov8233_start_play, stream %d\n", stream);
    
    audiov8233_fill_play_buffers(softstate);

    return AUDIO_SUCCESS;
}


static void
audiov8233_pause_play(audiohdl_t handle, int stream)
{
    struct audiov8233_softstate *softstate = audio_hlp_get_private(handle);
    uint8_t ctrl;

    DPRINTF4(AUDV8233M_DRIVER, "audiov8233_pause_play, stream %d\n", stream);

    mutex_enter(&softstate->intr_mutex);
    ctrl = V8233_GET8(V8233_RDCHN_REG(V8233_SGD_RDCHN_CTRL));
    V8233_PUT8(V8233_RDCHN_REG(V8233_SGD_RDCHN_CTRL), ctrl | V8233_SGD_CTRL_PAUSE);
    mutex_exit(&softstate->intr_mutex);
}


static void
audiov8233_stop_play(audiohdl_t handle, int stream)
{
    struct audiov8233_softstate *softstate = audio_hlp_get_private(handle);
    int i;
    struct sgd_chunk *chunk;
    sample_buffer_t *buf = &softstate->play_buf;

    DPRINTF4(AUDV8233M_DRIVER, "audiov8233_stop_play, stream %d\n", stream);

    if (softstate->play_timeout_id) {
	untimeout(softstate->play_timeout_id);
	softstate->play_timeout_id = 0;
    }

    mutex_enter(&softstate->intr_mutex);
    softstate->play_buf.io_started = FALSE;

    V8233_PUT8(V8233_RDCHN_REG(V8233_SGD_RDCHN_CTRL), V8233_SGD_CTRL_TERMINATE);

    /*
     * throw away the DMA chunks that were active on the VIA "PCM in"
     * DMA engine, release them to the free list
     */
    for (i = 0; i < V8233_NUM_SGD; i++) {
	if ((chunk = buf->active[i]) != NULL) {
	    buf->active[i] = NULL;

	    if (chunk->dma_bound) {
		ddi_dma_unbind_handle(chunk->dma_handle);
		chunk->dma_bound = FALSE;
	    }
	    chunk->next = buf->free;
	    buf->free = chunk;
	}
    }

    mutex_exit(&softstate->intr_mutex);
}


static int
audiov8233_get_delay(audiohdl_t handle, int stream)
{
    struct audiov8233_softstate *softstate = audio_hlp_get_private(handle);
    int samples;
    uint8_t idx;
    uint32_t cnt;
    
    DPRINTF4(AUDV8233M_DRIVER, "audiov8233_get_delay, stream %d\n", stream);

    mutex_enter(&softstate->intr_mutex);
    if (softstate->play_buf.io_started) {
	cnt = V8233_GET32(V8233_RDCHN_REG(V8233_SGD_RDCHN_CUR_CNT));
	idx = (cnt & V8233_SGD_CHN_CUR_IDX_MASK) >> V8233_SGD_CHN_CUR_IDX_SHIFT;
	cnt &= V8233_SGD_CHN_CUR_CNT_MASK;
	for (;;) {
	    idx = (idx+1) % V8233_NUM_SGD;
	    if (idx == softstate->play_buf.tail)
		break;
	    cnt += softstate->sgd->pcm_out_sgd[idx].bd_len;
	}

	samples = cnt;
	if (softstate->audio_info.play.channels == 2)   samples >>= 1;
	if (softstate->audio_info.play.precision == 16) samples >>= 1;
    } else
	samples = 0;
    mutex_exit(&softstate->intr_mutex);

    return samples;
}


static int
audiov8233_start_record(audiohdl_t handle, int stream)
{
    struct audiov8233_softstate *softstate = audio_hlp_get_private(handle);

    DPRINTF4(AUDV8233M_DRIVER, "audiov8233_start_record, stream %d\n", stream);

    audiov8233_prepare_record_buffers(softstate);

    return AUDIO_SUCCESS;
}


static void
audiov8233_stop_record(audiohdl_t handle, int stream)
{
    struct audiov8233_softstate *softstate = audio_hlp_get_private(handle);
    int i;
    struct sgd_chunk *chunk;
    sample_buffer_t *buf = &softstate->record_buf;

    DPRINTF4(AUDV8233M_DRIVER, "audiov8233_stop_record, stream %d\n", stream);
    mutex_enter(&softstate->intr_mutex);
    softstate->record_buf.io_started = FALSE;

    /* terminate SGD bus master */
    V8233_PUT8(V8233_WRCHN_REG(V8233_SGD_WRCHN_CTRL), V8233_SGD_CTRL_TERMINATE);

    /*
     * throw away the DMA chunks that were active on the VIA "PCM in"
     * DMA engine, release them to the free list
     */
    for (i = 0; i < V8233_NUM_SGD; i++) {
	if ((chunk = buf->active[i]) != NULL) {
	    buf->active[i] = NULL;

	    if (chunk->dma_bound) {
		ddi_dma_unbind_handle(chunk->dma_handle);
		chunk->dma_bound = FALSE;
	    }
	    chunk->next = buf->free;
	    buf->free = chunk;
	}
    }

    mutex_exit(&softstate->intr_mutex);
}


static int
audiov8233_ioctl(audiohdl_t handle, int channel,
		queue_t *q, mblk_t *mp, int *error)
{
    struct audiov8233_softstate *softstate = audio_hlp_get_private(handle);
    struct iocblk *iocp = (struct iocblk *)mp->b_rptr;
    mblk_t *mp2;
    caddr_t arg = 0;
    uint8_t vsr_vol;

    if (iocp->ioc_count == TRANSPARENT && (mp2 = mp->b_cont) != NULL)
	/* get the "arg" value from the streams msg */
	bcopy(mp2->b_rptr, &arg, sizeof(arg));

    DPRINTF4(AUDV8233M_DRIVER, "audiov8233_ioctl ioc_cmd=%x, arg=%p\n", iocp->ioc_cmd, arg);

    switch (iocp->ioc_cmd) {
    case AUDIOV8233_SET_VSR_VOL:
	if (softstate->playback_reg_base != V8233_SGD_RDCHN_IOBASE)
	    return AD_NACK;

	vsr_vol = (uint32_t) (long) arg;
	V8233_PUT8(V8233_RDCHN_REG(V8233_SGD_RDCHN_VOLL), vsr_vol);
	V8233_PUT8(V8233_RDCHN_REG(V8233_SGD_RDCHN_VOLR), vsr_vol);
	DPRINTF4(AUDV8233M_DRIVER, "VSR VOL=%x", vsr_vol);
	return AD_ACK;
    default:
	return AD_NACK;
    }
}


static int
audiov8233_iocdata(audiohdl_t handle, int channel,
		  queue_t *q, mblk_t *mp, int *error)
{
    struct copyresp *copyresp = (struct copyresp *)mp->b_rptr;
    mblk_t *data_mp;
    caddr_t arg = (caddr_t) copyresp->cp_private;

    if (mp->b_cont == NULL) return AD_NACK;
    data_mp = mp->b_cont;

    DPRINTF4(AUDV8233M_DRIVER, "audiov8233_iocdata cp_cmd=%x, arg=%p\n", copyresp->cp_cmd, arg);

    switch (copyresp->cp_cmd) {
    default:
	return AD_NACK;
    }
}


static audio_info_t audiov8233_defaults = {
    {
	8000,		     	/* samples per second */
	AUDIO_CHANNELS_MONO,	/* number of interleaved channels */
	AUDIO_PRECISION_8,	/* bit-width of each sample */
	AUDIO_ENCODING_ULAW,	/* data encoding method */

	AUDIO_MID_GAIN,		/* gain level: 0 - 255 */
	AUDIO_SPEAKER,		/* selected I/O port */
	AUDIO_SPEAKER | AUDIO_HEADPHONE | AUDIO_LINE_OUT,
#ifdef	AUDIOIO_HAVE_MODPORTS
	AUDIO_SPEAKER | AUDIO_HEADPHONE | AUDIO_LINE_OUT,
#endif
	0,			/* Reserved for future use */

	0,			/* I/O buffer size */

	0,			/* number of samples converted */
	0,			/* End Of File counter (play only) */

	0,
	0,
	0,
	AUDIO_MID_BALANCE,	/* stereo channel balance */
    },
    {
	8000,		     	/* samples per second */
	AUDIO_CHANNELS_MONO,	/* number of interleaved channels */
	AUDIO_PRECISION_8,	/* bit-width of each sample */
	AUDIO_ENCODING_ULAW,	/* data encoding method */

	AUDIO_MID_GAIN,		/* gain level: 0 - 255 */
	AUDIO_MICROPHONE,		/* selected I/O port */
	AUDIO_MICROPHONE | AUDIO_LINE_IN | AUDIO_CD |
		AUDIO_AUX1_IN | AUDIO_AUX2_IN | AUDIO_CODEC_LOOPB_IN,
#ifdef	AUDIOIO_HAVE_MODPORTS
	AUDIO_MICROPHONE | AUDIO_LINE_IN | AUDIO_CD |
		AUDIO_AUX1_IN | AUDIO_AUX2_IN | AUDIO_CODEC_LOOPB_IN,
#endif
	0,			/* Reserved for future use */

	0,			/* I/O buffer size */

	0,			/* number of samples converted */
	0,			/* End Of File counter (play only) */

	0,
	0,
	0,
	AUDIO_MID_BALANCE,	/* stereo channel balance */
    },
    AUDIO_MIN_GAIN,		/* monitor_gain */
    0,				/* output_muted */
#ifdef	AUDIOIO_HAS_REFCNT
    0,				/* driver reference count */
#endif
    {0, 0},			/* reserved fields */
#ifdef	AUDIOIO_HAS_FEATURES
    AUDIO_HWFEATURE_DUPLEX | AUDIO_HWFEATURE_IN2OUT | \
    	AUDIO_HWFEATURE_PLAY | AUDIO_HWFEATURE_RECORD,
    AUDIO_SWFEATURE_MIXER,
    0,
#endif
};


/*
 * Audio Driver ops vector definition
 */
static ah_ad_entry_t audiov8233_entry = {
    audiov8233_setup,
    audiov8233_teardown,
    audiov8233_set_config,
    audiov8233_set_format,
    audiov8233_start_play,
    audiov8233_pause_play,
    audiov8233_stop_play,
    audiov8233_start_record,
    audiov8233_stop_record,
    audiov8233_ioctl,
    audiov8233_iocdata,
    audiov8233_get_delay,
};


static audio_device_t audiov8233_dev_info = {
    "TOOLS,via8233",		/* name */
    "1.0",			/* version */
    "via8233"			/* config */
};


/* 0 term. list of sample rates (used as limits) */
static uint_t audiov8233_srs[] = {
    8000, 48000, 0
};

/* 0 terminated list of supported sample rates */
static uint_t audiov8233_srs_48k[] = {
    AC97_SAMPLE_RATE_48000, 0
};

static ah_ad_sample_rates_t audiov8233_sample_rates_48k = {
    AD_SRS_FLAG_SR_NOT_LIMITS, audiov8233_srs_48k
};

static uint_t audiov8233_chs[] = {
    AUDIO_CHANNELS_MONO,
    AUDIO_CHANNELS_STEREO,
    0,
};

static ah_ad_cap_comb_t audiov8233_cap_comb[] = {
    { AUDIO_PRECISION_16, AUDIO_ENCODING_LINEAR },  /* 16-bit signed PCM */
    { AUDIO_PRECISION_8,  AUDIO_ENCODING_LINEAR8 }, /* 8-bit unsigned PCM */
    { 0 },
};


static ah_ad_ch_cap_t audiov8233_ch_cap = {
    { AD_SRS_FLAG_SR_LIMITS, audiov8233_srs },
    audiov8233_chs,		/* list of channel types */
    0,				/* interrupt rate */
    2,				/* max channels */
    SAMPLEBUF_SIZE,		/* buffer size */
};
